#!/usr/bin/python
import getopt
import sys
import re

#
# MAIN PROGRAM
#

def simpleCPUv1a_as(argv):

  if len(sys.argv) <= 1:
    print ("Usage: simpleCPUv1a_as.py -i <input_file.asm>")
    print ("                          -o <output_file>") 
    print ("                          -a <address_offset>")
    print ("                          -t <input_file_type>")
    return

  # init variables #
  version = '1.1.1'
  source_filename = 'default.asm'
  tmp_filename = 'tmp.asm'
  word_filename = 'default.asc'
  mem_filename = 'default.mem'
  data_filename = 'default.dat' 

  address = 0
  byte_count = 0

  s_config = 'a:i:o:'
  l_config = ['address', 'input', 'output']

  input_file_present = False

  instruction_address = 0
  instruction_count   = 0

  label_dictionary = {}

  # capture commandline options #
  try:
    options, remainder = getopt.getopt(sys.argv[1:], s_config, l_config)
  except getopt.GetoptError as m:
    print "Error: ", m
    return

  # extract options #
  for opt, arg in options:
    if opt in ('-o', '--output'):
      if ".asc" in arg:
        word_filename = arg
      elif ".dat" in arg:
        data_filename = arg
      elif ".mem" in arg:
        mem_filename = arg
      else:
        word_filename = arg + ".asc"
        data_filename = arg + ".dat"
        mem_filename = arg + ".mem"
    elif opt in ('-i', '--input'):
      input_file_present = True
      if ".asm" in arg:
        source_filename = arg
      else:
        source_filename = arg + ".asm"
    elif opt in ('-a', '--address'):
      address = int(arg)
   
  # exit if no input file present # 
  if input_file_present:

    # open files #
    try:
      source_file = open(source_filename, "r")
    except IOError: 
      print "Error: Input file does not exist."
      return 

    try:
      word_file = open(word_filename, "w")
      mem_file = open(mem_filename, "w")
      data_file = open(data_filename, "w")
      tmp_file = open(tmp_filename, "w")
    except IOError: 
      print "Error: Could not open output files"
      return 

    # scan through code looking for labels

    instruction_address = address
    
    while True:
      line = source_file.readline()
      if line == '': 
        break
      if line[0] == '\r' or line[0] == '\n' or line[0] =='#':
        continue
      else:
        if ":" in line:
          key = re.sub(':\n', '', (line.replace(" ", "")).replace("\t", ""))
          label_dictionary[key] = instruction_address
        else:
          text = re.sub('\s+', ' ', line)
          words = text.split(' ')
          if words[1] != "":
            instruction_address += 1  

    #print label_dictionary
    source_file.close()     
    source_file = open(source_filename, "r")

    # generate tmp file with addresses for checking etc#

    instruction_address = address
    while True:
      line = source_file.readline()
      if line == '': 
        break
      if line[0] == '\r' or line[0] == '\n' or line[0] =='#':
        tmp_file.write( line )
      else:
        text = re.sub('\s+', ' ', line)
        words = text.split(' ')
          
        #print words  
        if ":" in text:
          continue
        elif words[1] == '':
          continue
        else:

          outputString = str.format('{:03}', instruction_address) + " "
          instruction_address += 1 

          for i in range(0, len(words)):
            if words[i] in label_dictionary:
              key = words[i]
              outputString = outputString + " " + str(label_dictionary[key])
            else:
              if words[i] != '':
                outputString = outputString + " " + words[i]

          outputString = outputString + "\n"
          tmp_file.write( outputString )

    # limit test #
    if address >  256:
      print "Error: program bigger than 256 instruction limit"
      return 

    tmp_file.close()

    # open TMP file #
    try:
      tmp_file = open(tmp_filename, "r")
    except IOError: 
      print "Error: could not output temp file"
      return 

    instruction_count = 0
    instruction_address = address

    # write start address to file #
    word_file.write(str.format('{:04X}', instruction_address) + ' ')

    while True:
      line = tmp_file.readline()
      if line == '':
        break 

      if line[0] =='\r' or line[0] =='\n' or line[0] =='#' or line[0] ==' ':
        pass
      else:
        text = re.sub('\s+', ' ', line.lower())
        words = text.split(' ')

        opcode = ''
        operand = ''
        
        if words[0].isdigit():

          # match opcode #
          if words[1]   == "move":
            opcode = "00 "
          elif words[1] == "add":
            opcode = "10 " 
          elif words[1] == "sub":
            opcode = "20 "
          elif words[1] == "and":
            opcode = "30 "
          elif words[1] == "load":
            opcode = "40 "
          elif words[1] == "store":
            if words[3] == '' or words[3] == '#':
              opcode = "50 "
            else:
              if int(words[2]) > 15:
                print "Error: invalid opcode" 
                print words 
                return
              else:
                opcode = "5" + str(words[2]) + " " 
          elif words[1] == "addm":
            opcode = "60 "
          elif words[1] == "subm":
            opcode = "70 "
          elif words[1] == "jump":
            opcode = "80 "
          elif words[1] == "jumpu":
            opcode = "80 "
          elif words[1] == "jumpz":
            opcode = "90 "
          elif words[1] == "jumpnz":
            opcode = "A0 "
          else:
            print "Error: invalid opcode" 
            print words 
            return

          length = 0 
          for i in range(len(words)):
            length = length + 1
            if words[i] == "#":
              break
            
          #print length
          if (length == 5):   
            data = words[3].rstrip()     
          elif (length >= 2) and (length < 5):
            data = words[2].rstrip()
          else:
            print "Error: invalid operand"
            print words 
            return

          if '0x' not in data:
            try:
              if int(data) < 256:
                operand = str.format('{:02X}', int(data)) + ' ' 
              else:
                print "Error: invalid operand (>255)"
                print words 
                return
            except ValueError:
                print "Error: invalid operand"
                print words 
                return
          else:
            if len(data) == 4:
              operand = data[2:4] + ' ' 
            elif len(data) == 3:
              operand = "0" + data[2] + ' ' 
            else:
              print "Error: invalid operand"
              print words 
              return


        # if opcode and operand good write instruction #
        #print opcode, operand
        if opcode == '' or operand == '':
          print "Error: invalid instruction"
          print words
          return
      
        else:
          data_file.write(str.format('{:04}', instruction_count) + ' ')
          bin_value = str.format('{:016b}', int((opcode.strip() + operand), 16)) 
          data_file.write( bin_value )
          data_file.write("\n")

          # update EPROM files #
          word_file.write(opcode.strip() + operand)

          # update mem file
          mem_file.write('@' + str.format('{:04X}', (instruction_count * 2)) + ' ')
          data_string = opcode.strip() + operand
          mem_file.write(data_string[3] + data_string[2] + data_string[1] + data_string[0] + "\n")

          instruction_count += 1
          byte_count += 1

          if byte_count == 16:
            byte_count = 0
            word_file.write("\n")

            instruction_address += 16
            addressString = str.format('{:04X}', instruction_address) + ' '
            word_file.write(addressString)

    # close files #
    source_file.close() 
    word_file.close()
    mem_file.close()
    data_file.close()
    tmp_file.close()

    # display info #
    outputString = "Number of instructions : " + str(instruction_count)
    print( outputString )

    if instruction_count > 0:
      max_address = instruction_count + address - 1
    outputString = "Address range : " + str(address) + " to " + str(max_address)
    print( outputString )

  else:
    print "Error: Input file not specified"
    return 

if __name__ == '__main__':
  simpleCPUv1a_as(sys.argv)

